/*
 * Ben NanoNote and EM027AS012 E-Paper Display Adapter Board communication.
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * SPI functions derived from those in lib/atben.c by Werner Almesberger:
 *
 * Copyright (C) 2010-2011 Werner Almesberger
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Driver-related code originally provided under the following terms:
 *
 * Copyright 2013 Pervasive Displays, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

#include <ubb/ubb.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include "bsp.h"
#include "Display_Hardware_Driver.h"

void bsp_init()
{
    OUT(UBB_DAT2);
    OUT(UBB_DAT3);
    OUT(UBB_CMD);
    OUT(UBB_CLK);
    OUT(UBB_DAT0);
    OUT(UBB_DAT1);
}

/* NanoNote-specific functions. */

void shift_select()
{
    CLR(MUX_S0);
}

void shift_update()
{
    SET(SHIFT_RCK);
    CLR(SHIFT_RCK);
}

void shift_init()
{
    /* Initiate a clear operation. */

    CLR(SHIFT_SCLR);

    /* Set clocks to known state. */

    CLR(SHIFT_SCK);
    CLR(SHIFT_RCK);

    /* Propagate the clear operation. */

    shift_update();

    /* Finalise the initialisation. */

    SET(SHIFT_SCLR);
}

void shift_reinit()
{
    /* Set clocks to known state. */

    CLR(SHIFT_SCK);
    CLR(SHIFT_RCK);

    /* Avoid clear operation. */

    SET(SHIFT_SCLR);
}

void shift_in(uint8_t v)
{
    if (v)
        SET(SHIFT_SI);
    else
        CLR(SHIFT_SI);

    SET(SHIFT_SCK);
    CLR(SHIFT_SCK);
}

void epd_select()
{
    SET(MUX_S0);
}

/* Callback functions. */

void bsp_InitDisplayHardware()
{
    /* This should do the following:

        PANEL = 0
        CS/SSEL = 0
        PWM = 0
        RESET = 0
        DIS = 0

        (Could shift 0, 0 -> PANEL, DIS
               or 0, 0, 0 -> PANEL, BORDER, DIS.)
    */

    shift_select();
    shift_init(); /* PANEL = DIS = 0 */

    epd_select();
    CLR(EPD_SSEL);
    CLR(EPD_PWM);
    CLR(EPD_RESET);
    CLR(EPD_MOSI);

    printf("Hardware initialised.\n");
}

void bsp_PowerOn()
{
    /* This should do the following:

        RESET = 0
        DIS = 0
        CS/SSEL = 0
    */

    epd_select();
    epd_rst_low();

    shift_select();
    shift_init(); /* PANEL = DIS = 0 */

    printf("RESET = PANEL = DIS = 0\n");

    /* Then, the following:

        PWM = 1/0 (5ms)
        PANEL = 1
        PWM = 1/0 (10ms)
        BORDER = 1?
        RESET = 1
        PWM = 1/0 (5ms)
        RESET = 0
        PWM = 1/0 (5ms)
        RESET = 1
        PWM = 1/0 (5ms)

        (Could shift 1 -> PANEL         => PANEL=1, DIS=0
               or 1, 1 -> PANEL, BORDER => PANEL=1, BORDER=1, DIS=0)
    */

    epd_select();
    epd_spi_init();
    printf("PWM at 5ms starting: %u\n", bsp_getMsTicks());
    epd_pwm_active(5);
    printf("PWM at 5ms done: %u\n", bsp_getMsTicks());

    shift_select();
    shift_init(); /* PANEL = DIS = 0 (still) */
    shift_in(1); /* PANEL = 1 */
    shift_update();
    printf("PANEL = 1\n");

    epd_select();
    printf("PWM at 10ms starting: %u\n", bsp_getMsTicks());
    epd_pwm_active(10);
    printf("PWM at 10ms done: %u\n", bsp_getMsTicks());
    epd_cs_high();
    epd_rst_high();
    printf("PWM at 5ms starting: %u\n", bsp_getMsTicks());
    epd_pwm_active(5);
    printf("PWM at 5ms done: %u\n", bsp_getMsTicks());
    epd_rst_low();
    printf("PWM at 5ms starting: %u\n", bsp_getMsTicks());
    epd_pwm_active(5);
    printf("PWM at 5ms done: %u\n", bsp_getMsTicks());
    epd_rst_high();
    printf("PWM at 5ms starting: %u\n", bsp_getMsTicks());
    epd_pwm_active(5);
    printf("PWM at 5ms done: %u\n", bsp_getMsTicks());

    printf("Hardware powered.\n");
}

void bsp_pwm_active(uint16_t delayInMs)
{
    uint16_t numOfIterations;

    numOfIterations = delayInMs * 100;
    for (; numOfIterations > 0; numOfIterations--)
    {
        // PWM_SET_HIGH;
        SET(EPD_PWM);
        bsp_delayUs(5);     //100kHz (96kHz ideal)
        // PWM_SET_LOW;
        CLR(EPD_PWM);
        bsp_delayUs(5);
    }
}

void bsp_Shutdown()
{
    /* This should complete the power off operation as follows:

        RESET = 0
        CS/SSEL = 0
        PANEL = 0
        BORDER = 0?
        DIS = 1
        DIS = 0 (after 150ms)

        (Could shift 0 -> PANEL         => PANEL=0, DIS=1
               or 0, 0 -> PANEL, BORDER => PANEL=0, BORDER=0, DIS=1)
    */

    epd_select();
    epd_rst_low();
    epd_cs_low();
    epd_spi_detach();

    shift_reinit();
    shift_select();
    shift_in(0); /* PANEL = 0; DIS = 1 */
    shift_update();
    printf("Wait for 150ms, starting: %u\n", bsp_getMsTicks());
    epd_delay_ms(150);
    printf("Wait for 150ms, done: %u\n", bsp_getMsTicks());
    shift_in(0); /* PANEL = 0; DIS = 0 */
    shift_update();

    printf("Hardware shut down.\n");
}

void bsp_spiInit()
{
    CLR(EPD_SCK);
}

/**
 * Send the given value via MOSI.
 */
void bsp_writeToDisplay(uint8_t *data, uint16_t len)
{
    uint8_t mask, *ptr, *limit = data + len;

    for (ptr = data; ptr < limit; ptr++)
    {
        for (mask = 0x80; mask; mask >>= 1)
        {
            if (*ptr & mask)
            {
                SET(EPD_MOSI);
            }
            else
            {
                CLR(EPD_MOSI);
            }

            SET(EPD_SCK);
            CLR(EPD_SCK);
        }
    }
}

void bsp_delayMs(uint32_t ms)
{
    bsp_delayUs(ms * 1000);
}

void bsp_delayUs(uint32_t us)
{
    #ifndef NANONOTE
    struct timespec tv;
    tv.tv_sec = 0;
    tv.tv_nsec = us * 100;
    nanosleep(&tv, NULL);
    #else
    uint64_t i = us * 16;
    while (i-- > 0);
    #endif
}

uint32_t bsp_getMsTicks()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

/* Unsupported. */

int16_t bsp_getTemp()
{
    return 20;
}
